@@ -1,11 +1,35 @@
 #!/usr/bin/env python3
 
 """
-Simple picture tool to replace JES picture tool
-CS Summer Practicum 2020
-Author: Gahngnin Kim
-Developed under the guidance of Dr. Jonathan Senning
-Modified by:
+pictureTool.py - script program to implement the "explore" functionality
+                 in JES for the JES4py package
+
+Written: 2020-07-24 Gahngnin Kim <gahngnin.kim@gordon.edu>
+Modified: 2020-08-08 Jonathan Senning <jonathan.senning@gordon.edu>
+- Completed work for crosshair cursor to work on Mac, Windows, and Linux (GTK)
+
+The "explore()" function in JES will open a new window and display the image
+imported from the given file path. The window provides an interactive picture
+tool for the user. It allows users to zoom in and out of the image and pick
+a pixel to examine its RGB values with the color preview. When a pixel is
+selected from the image, a crosshair will appear in that selected position.
+Unlike, "show()" function, it cannot be repainted with "repaint()" method.
+
+The JES4py's implementation of "explore()" function provides a nearly identical
+experience with an improvement compared to JES's. Display and image size
+detection code is added to ensure that the window size doesn't go beyond the
+display resolution even if the image is larger. The image panel will set
+scrollbars when a high-resolution image is imported will set the scrollbar to
+fit the image in the window. The program can also work standalone.
+
+This program was first developed during the Gordon College Computer Science
+Summer 2020 Practicum as a part of the JES4py project, under the guidance of
+Dr. Jonathan Senning.
+
+Summer 2020 JES4py Team: Dr. Jonathan Senning
+                         Nick Noormand
+                         Gahngnin Kim
+
 """
 
 import os
@@ -14,6 +38,114 @@ import wx
 import wx.lib.scrolledpanel
 # import wx.lib.inspection
 
+class Cursor:
+    width  = 7
+    height = 7
+    centerX = int((width - 1)/2)
+    centerY = int((height - 1)/2)
+    cursorBitmap = None
+    savedCursorPosition = None
+    cursorBackupBitmap = None
+
+    def __init__(self, width=7, height=7):
+        self.width = width
+        self.height = height
+        self.centerX = int((self.width - 1)/2)
+        self.centerY = int((self.height - 1)/2)
+
+        if wx.Platform != "__WXMAC__":
+            # WX on some macs does not support monochrome bitmaps which
+            # are necessary for bitmap masks.  Thus we don't need to do
+            # anything here
+            dc = wx.MemoryDC()
+
+            # Create the mask bitmap
+            cursorMask = wx.Bitmap(self.width, self.height, depth=1)
+            dc.SelectObject(cursorMask)
+            dc.SetPen(wx.Pen(wx.Colour(255, 255, 255), width=3))
+            dc.DrawLine(0, self.centerY, self.width-1, self.centerY)
+            dc.DrawLine(self.centerX, 0, self.centerX, self.height-1)
+
+            # Create the cursor bitmap
+            self.cursorBitmap = wx.Bitmap(self.width, self.height)
+            dc.SelectObject(self.cursorBitmap)
+            dc.SetPen(wx.Pen(wx.Colour(255, 255, 0)))
+            dc.DrawLine(0, self.centerY, self.width-1, self.centerY)
+            dc.DrawLine(self.centerX, 0, self.centerX, self.height-1)
+            self.cursorBitmap.SetMask(wx.Mask(cursorMask))
+
+            # done with dc
+            del dc
+
+    def drawCursor(self, canvas, x, y):
+        # Get scaled image size
+        W, H = canvas.image.GetSize()
+        W, H = int(W * canvas.zoomFactor), int(H * canvas.zoomFactor)
+
+        # Compute size and position of cursor and save corresponding portion
+        # of the image 
+        w, h = self.width, self.height
+        dx, dy = self.centerX, self.centerY
+        x0 = max(0, x-dx)
+        y0 = max(0, y-dy)
+        w0 = dx + min(dx+1, W-x)
+        h0 = dy + min(dy+1, H-y)
+        cursorRect = wx.Rect(x0, y0, w0, h0)
+        cursorPosition = x-dx, y-dy
+        self.savedCursorPosition = x0, y0
+        self.cursorBackupBitmap = canvas.bmp.GetSubBitmap(cursorRect)
+        # print(f"Cursor size: {self.cursorSize}, Cursor Rect: {cursorRect}")
+
+        # Draw the cursor - use bitmap on Windows and GTK, DC Draw on Mac
+        if wx.Platform == "__WXMSW__":
+            dc = wx.ClientDC(canvas.imageCtrl)
+        else:
+            dc = wx.ClientDC(canvas.imagePanel)
+            canvas.imagePanel.DoPrepareDC(dc)
+        dc.SetClippingRegion(0, 0, W, H)
+        if self.cursorBitmap is None:
+            # We need to manually draw the cursor
+            dx, dy = self.centerX-1, self.centerY-1
+            dc.SetPen(wx.Pen(wx.Colour(0, 0, 0), width=3))
+            dc.DrawLine(x-dx, y, x+dx, y)
+            dc.DrawLine(x, y-dy, x, y+dy)
+            dc.SetPen(wx.Pen(wx.Colour(255, 255, 0), width=1))
+            dc.DrawLine(x-dx, y, x+dx, y)
+            dc.DrawLine(x, y-dy, x, y+dy)
+        else:
+            # Display cursor bitmap
+            dc.DrawBitmap(self.cursorBitmap, cursorPosition, True)
+        del dc
+
+    def undrawPreviousCursor(self, canvas):
+        """Restore bitmap (if any) saved from previous cursor event
+        """
+        if self.cursorBackupBitmap is not None:
+            if wx.Platform == "__WXMSW__":
+                dc = wx.ClientDC(canvas.imageCtrl)
+            else:
+                dc = wx.ClientDC(canvas.imagePanel)
+                canvas.imagePanel.DoPrepareDC(dc)
+            dc.DrawBitmap(self.cursorBackupBitmap, self.savedCursorPosition, False)
+            del dc
+
+    def drawCrosshairs(self, canvas):
+        """Draw image with crosshairs to indicate selected position
+        """
+        # Get coordinates of current cursor location
+        x = int(int(canvas.pixelTxtX.GetValue()) * canvas.zoomFactor)
+        y = int(int(canvas.pixelTxtY.GetValue()) * canvas.zoomFactor)
+
+        # Restore bitmap (if any) saved from previous cursor event
+        # and draw new cursor
+        self.undrawPreviousCursor(canvas)
+        self.drawCursor(canvas, x, y)
+
+    def clearBackupBitmap(self):
+        """Forget saved bitmap buffer
+        """
+        if self.cursorBackupBitmap is not None:
+            self.cursorBackupBitmap = None
 
 class MainWindow(wx.Frame):
 
@@ -22,15 +154,16 @@ class MainWindow(wx.Frame):
     ColorPanelHeight = 70
     PaintChipSize = 24
     zoomLevels = [25, 50, 75, 100, 150, 200, 500]
-    zoomLevel = float(zoomLevels[3]) / 100.0
+    zoomFactor = float(zoomLevels[3]) / 100.0
+    cursorSize = (7, 7)
+    cursorBitmap = None
     x = 0
     y = 0
 
     def __init__(self, filename, parent, title):
         # Load image and get image size
         self.image = wx.Image(filename, wx.BITMAP_TYPE_ANY)
-        self.bmp = wx.Bitmap(self.image, wx.BITMAP_TYPE_ANY)
-
+        self.bmp = wx.Bitmap(self.image)
         super(MainWindow, self).__init__(parent=parent, title=title, style=wx.DEFAULT_FRAME_STYLE)
 
         self.InitUI()
@@ -39,36 +172,33 @@ class MainWindow(wx.Frame):
         # wx.lib.inspection.InspectionTool().Show() # Inspection tool for debugging
 
     def InitUI(self):
-        wScr, hScr = wx.DisplaySize()
-        wView, hView = int(wScr * 0.95), int(hScr * 0.85)
-        w, h = self.image.GetSize()
+        # Get display size and reduce width and height by small percentage
+        wView, hView = wx.DisplaySize()
+        wView, hView = int(wView * 0.95), int(hView * 0.85)
 
-        # w = min(max(self.MinWindowWidth, w), wView)
-        # h = min(max(self.MinWindowHeight, h + self.ColorPanelHeight), hView)
+        # Get image dimensions and adjust window size to hold as much of the image
+        # as possible but stil be within display
+        w, h = self.image.GetSize()
         w = min(max(self.MinWindowWidth, w), wView)
         h = min(max(self.MinWindowHeight, h + self.ColorPanelHeight), hView)
 
-        # setup the zoom menu
+        # Setup the zoom menu
         self.setupZoomMenu()
 
-        # create main sizer
+        # Create main sizer
         self.mainSizer = wx.BoxSizer(wx.VERTICAL)
 
-        # setup the color information panel
+        # Setup the color information panel
         self.setupColorInfoDisplay()
         self.mainSizer.Add(self.colorInfoPanel, 0, wx.EXPAND|wx.ALL, 0)
 
-        # set up the image display panel
+        # Set up the image display panel
         self.setupImageDisplay()
         self.mainSizer.Add(self.imagePanel, 0, wx.EXPAND|wx.ALL, 0)
-
         self.SetSizer(self.mainSizer)
-        # self.SetSize((w, h))
         self.Fit()
-        self.imagePanel.SetupScrolling(scrollToTop=True)
 
-        # w, h = self.image.GetSize()
-        # h = h + self.ColorPanelHeight
+        # Set window sizes
         self.SetSize((w, h))
         self.SetClientSize((w,h))
 
@@ -81,24 +211,7 @@ class MainWindow(wx.Frame):
             zoomID = zoomMenu.Append(wx.ID_ANY, item, helpString)
             self.Bind(wx.EVT_MENU, self.onZoom, zoomID)
 
-        # menuZoom25 = zoomMenu.Append(wx.ID_ANY, "25%","Zoom by 25%")
-        # menuZoom50 = zoomMenu.Append(wx.ID_ANY, "50%","Zoom by 50%")
-        # menuZoom75 = zoomMenu.Append(wx.ID_ANY, "75%","Zoom by 75%")
-        # menuZoom100 = zoomMenu.Append(wx.ID_ANY, "100%","Zoom by 100%")
-        # menuZoom150 = zoomMenu.Append(wx.ID_ANY, "150%","Zoom by 150%")
-        # menuZoom200 = zoomMenu.Append(wx.ID_ANY, "200%","Zoom by 200%")
-        # menuZoom500 = zoomMenu.Append(wx.ID_ANY, "500%","Zoom by 500%")
-
-        # # Set menu events
-        # self.Bind(wx.EVT_MENU, self.onZoom, menuZoom25)
-        # self.Bind(wx.EVT_MENU, self.onZoom, menuZoom50)
-        # self.Bind(wx.EVT_MENU, self.onZoom, menuZoom75)
-        # self.Bind(wx.EVT_MENU, self.onZoom, menuZoom100)
-        # self.Bind(wx.EVT_MENU, self.onZoom, menuZoom150)
-        # self.Bind(wx.EVT_MENU, self.onZoom, menuZoom200)
-        # self.Bind(wx.EVT_MENU, self.onZoom, menuZoom500)
-        
-        # Creating the menubar.
+        # Create the menubar
         menuBar = wx.MenuBar()
         menuBar.Append(zoomMenu, "&Zoom") # Adds the "zoomMenu" to the MenuBar
         self.SetMenuBar(menuBar) # Adds the MenuBar to the Frame content.
@@ -142,24 +255,19 @@ class MainWindow(wx.Frame):
 
         # Textboxes to display X and Y coordinates on click
         self.pixelTxtX = wx.TextCtrl(self.colorInfoPanel, wx.ALIGN_CENTER, \
-            style=wx.TE_PROCESS_ENTER, size=(50,-1))
+            style=wx.TE_PROCESS_ENTER, size=(50, -1))
         self.pixelTxtY = wx.TextCtrl(self.colorInfoPanel, wx.ALIGN_CENTER, \
-            style=wx.TE_PROCESS_ENTER, size=(50,-1))
+            style=wx.TE_PROCESS_ENTER, size=(50, -1))
         self.pixelTxtX.Bind(wx.EVT_TEXT_ENTER, self.ImageCtrl_OnEnter)
         self.pixelTxtY.Bind(wx.EVT_TEXT_ENTER, self.ImageCtrl_OnEnter)
 
         # Add first row components to sizer
-        # X coordinate display
-        sizer1.Add(lblX, 0, flag=wx.CENTER, border=0)
+        sizer1.Add(lblX, 0, flag=wx.CENTER, border=0) # X coordinate display
         sizer1.Add(buttonX_L, 0, border=0)
         sizer1.Add(self.pixelTxtX, 0, flag=wx.CENTER, border=5) # Text Control Box
         sizer1.Add(buttonX_R, 0, border=0)
-
-        # Horizonal spacer
-        sizer1.Add((5, -1))
-
-        # Y coordinate display
-        sizer1.Add(lblY, 0, flag=wx.CENTER, border=0)
+        sizer1.Add((5, -1)) # Horizonal spacer
+        sizer1.Add(lblY, 0, flag=wx.CENTER, border=0) # Y coordinate display
         sizer1.Add(buttonY_L, 0, border=0)
         sizer1.Add(self.pixelTxtY, 0, flag=wx.CENTER, border=5) # Text Control Box
         sizer1.Add(buttonY_R, 0, border=0)
@@ -167,9 +275,7 @@ class MainWindow(wx.Frame):
         # Add first row sizer to the main sizer
         mainSizer.Add(sizer1, 0, flag=wx.LEFT|wx.RIGHT|
                     wx.TOP|wx.ALIGN_CENTER_HORIZONTAL|wx.ALL, border=1)
-
-        # Vertical spacer
-        mainSizer.Add((-1, 5))
+        mainSizer.Add((-1, 5)) # Vertical spacer
 
         # ---- Second row ----
 
@@ -189,23 +295,16 @@ class MainWindow(wx.Frame):
 
         # Add items to the second sizer (sizer2)
         sizer2.Add(self.rgbValue, 0, flag=wx.ALIGN_CENTER, border=5)
-
-        # Horizontal spacer
-        sizer2.Add((10, -1), 0)
+        sizer2.Add((10, -1), 0) # Horizontal spacer
 
         # Small image that shows the color at the selected pixel
         sizer2.Add(self.colorPreview, 0, flag=wx.ALIGN_CENTER, border=5) 
 
         # Add sizer2 to the main sizer
         mainSizer.Add(sizer2, 0, flag=wx.ALIGN_CENTER_HORIZONTAL|wx.ALL, border=1)
-
-        # Vertical spacer
-        mainSizer.Add((-1, 5))
+        mainSizer.Add((-1, 5)) # Vertical spacer
 
         self.colorInfoPanel.SetSizer(mainSizer)
-        #mainSizer.Fit(panel)
-        #panel.Layout()
-        #self.Show()
 
     def setupImageDisplay(self):
         """Set up image display panel
@@ -224,20 +323,19 @@ class MainWindow(wx.Frame):
         if wx.Platform == "__WXMSW__" or wx.Platform == "__WXGTK__":
             self.imageCtrl.Bind(wx.EVT_LEFT_DOWN, self.ImageCtrl_OnMouseClick)
             self.imageCtrl.Bind(wx.EVT_MOTION, self.ImageCtrl_OnMouseClick)
-            #self.imageCtrl.Bind(wx.EVT_LEFT_UP, self.OnPaint)
         elif wx.Platform == "__WXMAC__":
             self.imagePanel.Bind(wx.EVT_LEFT_DOWN, self.ImageCtrl_OnMouseClick)
             self.imagePanel.Bind(wx.EVT_MOTION, self.ImageCtrl_OnMouseClick) 
-            #self.imagePanel.Bind(wx.EVT_LEFT_UP, self.OnPaint)
-        #panel.SetFocus()
-        #panel.Bind(wx.EVT_LEFT_DOWN, self.onFocus)
 
+        # Add image control to image panel
         mainSizer = wx.BoxSizer(wx.VERTICAL)
         mainSizer.Add(self.imageCtrl, 0, wx.ALIGN_LEFT|wx.ALIGN_TOP|wx.ALL, 0)
         self.imagePanel.SetSizer(mainSizer)
         self.mainSizer.Fit(self.imagePanel)
         self.imagePanel.SetupScrolling(scrollToTop=True)
-        #return panel
+
+        # Create cursor
+        self.crosshair = Cursor()
 
     def clipOnBoundary(self):
         """Clips x and y to be valid pixel coordinates
@@ -252,20 +350,15 @@ class MainWindow(wx.Frame):
         self.pixelTxtX, self.pixelTxtY : wx.TextCtrl
             the textboxes displaying the pixel coordinates
         """
-        imageSize = self.image.GetSize()
-        #print('image size: ', imageSize)
+        width, height = self.image.GetSize()
         if self.x < 0:
             self.x = 0
-            self.y = 0
-        elif self.x >= imageSize[0]:
-            self.x = imageSize[0] - 1
-            self.y = imageSize[1] - 1
+        elif self.x >= width:
+            self.x = width - 1
         if self.y < 0:
-            self.x = 0
             self.y = 0
-        elif self.y >= imageSize[1]:
-            self.y = imageSize[1] - 1
-            self.x = imageSize[0] - 1
+        elif self.y >= height:
+            self.y = height - 1
         self.pixelTxtX.SetValue(str(self.x))
         self.pixelTxtY.SetValue(str(self.y))
         self.updateColorInfo()
@@ -286,10 +379,8 @@ class MainWindow(wx.Frame):
         bmp = wx.Bitmap(self.PaintChipSize, self.PaintChipSize)
         dc = wx.MemoryDC()
         dc.SelectObject(bmp)
-        pen = wx.Pen(wx.Colour(0,0,0)) # black for outline
-        dc.SetPen(pen)
-        brush = wx.Brush(wx.Colour(r,g,b)) # current image pixel color
-        dc.SetBrush(brush)
+        dc.SetPen(wx.Pen(wx.Colour(0,0,0))) # black for outline
+        dc.SetBrush(wx.Brush(wx.Colour(r,g,b))) # current image pixel color
         dc.DrawRectangle(0, 0, self.PaintChipSize, self.PaintChipSize)
         del dc
         self.colorPreview.SetBitmap(bmp) # Replace the bitmap with the new one
@@ -298,30 +389,22 @@ class MainWindow(wx.Frame):
     def updateView(self):
         """Scale image according to the zoom factor and (re)display it
         """
-        imageSize = self.image.GetSize()
-        w = int(imageSize[0] * self.zoomLevel)
-        h = int(imageSize[1] * self.zoomLevel)
+        w, h = self.image.GetSize()
+        w, h = int(w * self.zoomFactor), int(h * self.zoomFactor)
         image = self.image.Scale(w, h)
-        self.bmp = wx.Bitmap(image, wx.BITMAP_TYPE_ANY)
+        self.bmp = wx.Bitmap(image)
         self.imageCtrl.SetBitmap(self.bmp)
 
+        # event = wx.PyCommandEvent(wx.EVT_TEXT_ENTER.typeId, self.GetId())
+        # wx.PostEvent(self.GetEventHandler(), event)
+
+        # Forget any saved bitmap buffer
+        self.crosshair.clearBackupBitmap()
+        
     def drawCrosshairs(self):
-        """This feature works fine with Linux but not with Windows
+        """Draw image with crosshairs to indicate selected position
         """
-        pass
-        # """Draw image with crosshairs to indicate selected position
-        # """
-        # dc = wx.ClientDC(self.imageCtrl)
-        # self.imagePanel.DoPrepareDC(dc)
-        # origin = dc.GetDeviceOrigin()
-        # scrolledPosition = self.imagePanel.CalcUnscrolledPosition(origin)
-        # x = int(self.x * self.zoomLevel) + scrolledPosition[0]
-        # y = int(self.y * self.zoomLevel) + scrolledPosition[1]
-        # print (origin, scrolledPosition)
-        # print (x, y)
-        # dc.DrawBitmap(self.bmp, scrolledPosition, False)
-        # dc.SetPen(wx.Pen(wx.Colour(0, 0, 0), 1, wx.DOT))
-        # dc.CrossHair(x, y)
+        self.crosshair.drawCrosshairs(self)
 
 # ===========================================================================
 # Event handlers
@@ -332,15 +415,15 @@ class MainWindow(wx.Frame):
 
     def onZoom(self, event):
         """Sets desired zoom level and updates image display
-
-        (Zoom event handler)
         """
         id_selected = event.GetId() # Gets the event id of the selected menu item
         obj = event.GetEventObject() # Gets the event object
         menuItem = obj.GetLabelText(id_selected) # Gets the label text of the menu item
-        self.zoomLevel = float(menuItem.replace('%','')) / 100.0
-        self.x = self.y = 0
-        self.imagePanel.Scroll(self.x, self.y) # "non-scrolled" position
+        self.zoomFactor = float(menuItem.replace('%','')) / 100.0
+
+        # Reset position of panel
+        self.imagePanel.Scroll(0, 0) # "non-scrolled" position
+
         self.PostSizeEvent()
         self.clipOnBoundary()
         self.updateView()
@@ -373,16 +456,18 @@ class MainWindow(wx.Frame):
         """
         if event.LeftIsDown():
             event.Skip()
-            dc = wx.ClientDC(self)
-            self.imagePanel.DoPrepareDC(dc)
             if wx.Platform == "__WXMSW__":
-                dc_pos = event.GetPosition()
+                cursorPosition = event.GetPosition()
             elif wx.Platform == "__WXGTK__" or wx.Platform == "__WXMAC__":
-                dc_pos = event.GetLogicalPosition(dc)
-            self.coordinates = dc_pos
-            del dc
-            self.x = int(dc_pos.x / self.zoomLevel)
-            self.y = int(dc_pos.y / self.zoomLevel)
+                dc = wx.ClientDC(self)
+                self.imagePanel.DoPrepareDC(dc)
+                cursorPosition = event.GetLogicalPosition(dc)
+                del dc
+
+            self.x = int(cursorPosition.x / self.zoomFactor)
+            self.y = int(cursorPosition.y / self.zoomFactor)
+            # print(f"\x1b[1mcursorPosition = {self.cursorPosition}\x1b[0m")
+
             self.clipOnBoundary()
             self.drawCrosshairs()
 
